/*
	This file is part of lgui.

	lgui is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	lgui is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with lgui.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2008 Lucas Hermann Negri
*/

#include "treemodel.h"

static gboolean priv_foreach_callback(GtkTreeModel* model, GtkTreePath* path, 
	GtkTreeIter* iter, lua_State* L)
{
	int top = lua_gettop(L);
	
	/* Push the function and userdata */
	lua_pushvalue(L, -2);
	lua_pushvalue(L, -2);
	
	/* Push the iters */
	struct_new(L, path, FALSE);
	struct_new(L, iter, FALSE);

	/* Call the function */
	lua_call(L, 3, 1);

	/* Get the return */
	gboolean res = lua_toboolean(L, -1);
	
	lua_settop(L, top);

	#ifdef IDEBUG
	g_debug("ForeachCallback returning!");
	#endif

	return res;
}

/* GtkTreeIter */

int lgui_tree_iter_new(lua_State* L)
{
	GtkTreeIter* ptr = g_malloc(sizeof(GtkTreeIter));
	struct_new(L, ptr, TRUE);

	return 1;
}

int lgui_tree_iter_copy(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	
	GtkTreeIter* ptr = g_malloc(sizeof(GtkTreeIter));
	*ptr = *((GtkTreeIter*)obj->pointer);
	struct_new(L, ptr, TRUE);

	return 1;
}

/* GtkTreePath */

static int priv_tree_path_gc(lua_State* L)
{
	#ifdef IDEBUG
	g_debug("Garbage collecting a GtkTreePath!");
	#endif
	
	Object* obj = lua_touserdata(L, 1);
	gtk_tree_path_free(obj->pointer);
	
	return 0;
}

static int priv_tree_path_tostring(lua_State* L)
{
	Object* obj = lua_touserdata(L, 1);
	
	gchar* str = gtk_tree_path_to_string(obj->pointer);
	
	if(str)
	{
		lua_pushstring(L, str);
		g_free(str);
	}
	else
		lua_pushstring(L, "");
	
	return 1;
}

static void priv_push_tree_path(lua_State* L, GtkTreePath* path)
{
	if(path)
	{
		#ifdef IDEBUG
		g_debug("Pushing a GtkTreePath!");
		#endif

		Object* obj = lua_newuserdata(L, sizeof(Object));
		obj->pointer = path;
		obj->need_unref = TRUE; /* Not used, just for consistency */
		int top = lua_gettop(L);
		
		/* Shared metatable */
		if(luaL_newmetatable(L, "__lguiTrEePaThMt"))
		{
			lua_getfield(L, LUA_GLOBALSINDEX, "__lobjRegisterSpecial");
			lua_pushvalue(L, -2);
			lua_pushliteral(L, "lgui");
			lua_pushliteral(L, "TreePath");
			lua_pushcfunction(L, priv_tree_path_gc);
			lua_pushcfunction(L, priv_tree_path_tostring);
			lua_call(L, 5, 0);
		}

		/* Set the object metatable */
		lua_setmetatable(L, top);
	}
	else
		lua_pushnil(L);
}

int lgui_tree_path_new(lua_State* L)
{
	GtkTreePath* path = gtk_tree_path_new();
	priv_push_tree_path(L, path);
	
	return 1;
}

int lgui_tree_path_new_first(lua_State* L)
{
	GtkTreePath* path = gtk_tree_path_new_first();
	priv_push_tree_path(L, path);
	
	return 1;
}

int lgui_tree_path_new_from_string(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	GtkTreePath* path = gtk_tree_path_new_from_string(lua_tostring(L, 1));
	priv_push_tree_path(L, path);
	
	return 1;
}

int lgui_tree_path_copy(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	
	GtkTreePath* path = gtk_tree_path_copy(obj->pointer);
	priv_push_tree_path(L, path);
	
	return 1;
}

int lgui_tree_path_compare(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);
	
	Object* obj1 = lua_touserdata(L, 1);
	Object* obj2 = lua_touserdata(L, 2);
	
	gint res = gtk_tree_path_compare(obj1->pointer, obj2->pointer);
	lua_pushinteger(L, res);
	
	return 1;
}

int lgui_tree_path_append_index(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	
	Object* obj = lua_touserdata(L, 1);
	gtk_tree_path_append_index(obj->pointer, lua_tointeger(L, 2));
	
	return 0;
}

int lgui_tree_path_prepend_index(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	
	Object* obj = lua_touserdata(L, 1);
	gtk_tree_path_prepend_index(obj->pointer, lua_tointeger(L, 2));
	
	return 0;
}

int lgui_tree_path_get_depth(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	
	Object* obj = lua_touserdata(L, 1);
	gint depth = gtk_tree_path_get_depth(obj->pointer);
	lua_pushinteger(L, depth);
	
	return 1;
}

int lgui_tree_path_get_indices(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	
	Object* obj = lua_touserdata(L, 1);
	gint* indices = gtk_tree_path_get_indices(obj->pointer);

	if(indices)
	{
		gint depth = gtk_tree_path_get_depth(obj->pointer), i;
		lua_newtable(L);
		
		for(i = 0; i < depth; ++i)
		{
			lua_pushinteger(L, indices[i]);
			lua_rawseti(L, -2, i + 1);
		}
	}
	else
		lua_pushnil(L);
	
	return 1;
}

int lgui_tree_path_next(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	
	gtk_tree_path_next(obj->pointer);
	
	return 0;
}

int lgui_tree_path_prev(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	
	gboolean res = gtk_tree_path_prev(obj->pointer);
	lua_pushboolean(L, res);
	
	return 1;
}

int lgui_tree_path_up(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	
	gboolean res = gtk_tree_path_up(obj->pointer);
	lua_pushboolean(L, res);
	
	return 1;
}

int lgui_tree_path_down(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* obj = lua_touserdata(L, 1);
	
	gtk_tree_path_down(obj->pointer);
	
	return 0;
}

int lgui_tree_path_is_ancestor(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);
	
	Object* obj1 = lua_touserdata(L, 1);
	Object* obj2 = lua_touserdata(L, 2);
	
	gboolean res = gtk_tree_path_is_ancestor(obj1->pointer, obj2->pointer);
	lua_pushboolean(L, res);
	
	return 1;
}

int lgui_tree_path_is_descendant(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);
	
	Object* obj1 = lua_touserdata(L, 1);
	Object* obj2 = lua_touserdata(L, 2);
	
	gboolean res = gtk_tree_path_is_descendant(obj1->pointer, obj2->pointer);
	lua_pushboolean(L, res);
	
	return 1;
}

/* GtkTreeModel */

int lgui_tree_model_get_iter_from_string(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);
	luaL_checktype(L, 3, LUA_TSTRING);

	Object* obj = lua_touserdata(L, 1);
	Object* stc = lua_touserdata(L, 2);

	gboolean res = gtk_tree_model_get_iter_from_string(GTK_TREE_MODEL(obj->pointer),
		stc->pointer, lua_tostring(L, 3));

	lua_pushboolean(L, res);

	return 1;
}

int lgui_tree_model_get_string_from_iter(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);

	Object* obj = lua_touserdata(L, 1);
	Object* stc = lua_touserdata(L, 2);
	gchar* str = gtk_tree_model_get_string_from_iter(GTK_TREE_MODEL(obj->pointer),
		stc->pointer);
	lua_pushstring(L, str);
	g_free(str);

	return 1;
}

int lgui_tree_model_get_iter(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);
	luaL_checktype(L, 3, LUA_TUSERDATA);

	Object* obj1 = lua_touserdata(L, 1);
	Object* obj2 = lua_touserdata(L, 2);
	Object* obj3 = lua_touserdata(L, 3);

	gboolean res = gtk_tree_model_get_iter(GTK_TREE_MODEL(obj1->pointer), 
		obj2->pointer, obj3->pointer);

	lua_pushboolean(L, res);

	return 1;
}

int lgui_tree_model_get_iter_first(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);

	Object* obj = lua_touserdata(L, 1);
	Object* stc = lua_touserdata(L, 2);

	gboolean res = gtk_tree_model_get_iter_first(GTK_TREE_MODEL(obj->pointer),
	(GtkTreeIter*)stc->pointer);

	lua_pushboolean(L, res);

	return 1;
}

int lgui_tree_model_iter_next(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);

	Object* obj = lua_touserdata(L, 1);
	Object* stc = lua_touserdata(L, 2);

	gboolean res = gtk_tree_model_iter_next(GTK_TREE_MODEL(obj->pointer),
	(GtkTreeIter*)stc->pointer);

	lua_pushboolean(L, res);

	return 1;
}

int lgui_tree_model_get(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);

	int size = lua_gettop(L), i;
	GtkTreeModel* model = GTK_TREE_MODEL(((Object*)lua_touserdata(L, 1))->pointer);
	GtkTreeIter*  iter  = ((Object*)lua_touserdata(L, 2))->pointer;
	lua_getfield(L, LUA_GLOBALSINDEX, "__lobjValuePush");

	/* Get field by field */
	for(i = 3; i <= size; ++i)
	{
		luaL_checktype(L, i, LUA_TNUMBER);
		int column = lua_tointeger(L, i);

		#ifdef IDEBUG
		GType type = gtk_tree_model_get_column_type(model, column);
		g_debug("Getting listModel column %i, of type %s!", column, g_type_name(type));
		#endif
		
		GValue val = {0};
		gtk_tree_model_get_value(model, iter, column, &val);
		lua_pushvalue(L, size + 1);
		lua_pushlightuserdata(L, &val);
		lua_call(L, 1, 1);
		g_value_unset(&val);
	}

	return size - 2;
}

int lgui_tree_model_get_n_columns(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);

	Object* obj = lua_touserdata(L, 1);
	gint cols = gtk_tree_model_get_n_columns(GTK_TREE_MODEL(obj->pointer));
	lua_pushnumber(L, cols);

	return 1;
}

int lgui_tree_model_foreach(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TFUNCTION);
	
	Object* obj = lua_touserdata(L, 1);
	lua_pushvalue(L, 2);
	lua_pushvalue(L, 3);

	gtk_tree_model_foreach(GTK_TREE_MODEL(obj->pointer), 
		(GtkTreeModelForeachFunc)priv_foreach_callback, L);

	return 0;
}

int lgui_tree_model_get_path(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);
	
	Object* obj1 = lua_touserdata(L, 1);
	Object* obj2 = lua_touserdata(L, 2);
	
	GtkTreePath* path = gtk_tree_model_get_path(GTK_TREE_MODEL(obj1->pointer),
		obj2->pointer);
		
	priv_push_tree_path(L, path);
	return 1;
}

int lgui_tree_model_row_changed(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);
	luaL_checktype(L, 3, LUA_TUSERDATA);
	
	Object* obj1 = lua_touserdata(L, 1);
	Object* obj2 = lua_touserdata(L, 2);
	Object* obj3 = lua_touserdata(L, 3);
	
	gtk_tree_model_row_changed(GTK_TREE_MODEL(obj1->pointer), obj2->pointer,
		obj3->pointer);
		
	return 0;
}

int lgui_tree_model_row_inserted(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);
	luaL_checktype(L, 3, LUA_TUSERDATA);
	
	Object* obj1 = lua_touserdata(L, 1);
	Object* obj2 = lua_touserdata(L, 2);
	Object* obj3 = lua_touserdata(L, 3);
	
	gtk_tree_model_row_inserted(GTK_TREE_MODEL(obj1->pointer), obj2->pointer,
		obj3->pointer);
		
	return 0;
}

int lgui_tree_model_row_has_child_toggled(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);
	luaL_checktype(L, 3, LUA_TUSERDATA);
	
	Object* obj1 = lua_touserdata(L, 1);
	Object* obj2 = lua_touserdata(L, 2);
	Object* obj3 = lua_touserdata(L, 3);
	
	gtk_tree_model_row_has_child_toggled(GTK_TREE_MODEL(obj1->pointer),
		obj2->pointer, obj3->pointer);
		
	return 0;
}

int lgui_tree_model_row_deleted(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);
	
	Object* obj1 = lua_touserdata(L, 1);
	Object* obj2 = lua_touserdata(L, 2);
	
	gtk_tree_model_row_deleted(GTK_TREE_MODEL(obj1->pointer), obj2->pointer);
		
	return 0;
}

int lgui_tree_model_rows_reordered(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);
	luaL_checktype(L, 4, LUA_TTABLE);
	
	Object* obj1 = lua_touserdata(L, 1);
	Object* obj2 = lua_touserdata(L, 2);
	GtkTreeIter* iter;
	
	gint size = lua_objlen(L, 4), i;
	gint* new_order = g_malloc(sizeof(gint) * sizeof(gint));
	
	for(i = 1; i <= size; ++i)
	{
		lua_rawgeti(L, 4, i);
		new_order[i - 1] = lua_tointeger(L, -1);
		lua_pop(L, 1);
	}
	
	if(lua_type(L, 3) == LUA_TUSERDATA)
		iter = ((Object*)lua_touserdata(L, 3))->pointer;
	else
		iter = NULL;
	
	gtk_tree_model_rows_reordered(GTK_TREE_MODEL(obj1->pointer),
		obj2->pointer, iter, new_order);
		
	g_free(new_order);
		
	return 0;
}
